"""
    Seed the database with repositories.
"""
from datoso.database.models import Seed

repositories = [
    {
        "name": "No-Intro",
        "short_name": "nointro",
        "description": "No-Intro is a collection of DATs aimed for video game preservation especialized in cartridge-based and digital systems.",
        "url": "https://wiki.no-intro.org/index.php",
        "icon": "https://www.no-intro.org/favicon.ico",
        "update_script": "lib/nointro/update.py",
        "type": "system"
    },
    {
        "name": "Redump",
        "short_name": "redump",
        "description": "Redump.org is a disc preservation database and internet community dedicated to collecting precise and accurate information about every video game ever released on optical media of any system.",
        "url": "http://redump.org/",
        "icon": "http://redump.org/favicon.ico",
        "update_script": "lib/redump/update",
        "type": "system"
    },
    {
        "name": "Translated-English",
        "short_name": "t_en",
        "description": "Translated-English is a collection of DATs aimed for video game translations.",
        "url": "http://archive.org/En-ROMs/",
        "icon": "http://archive.org/favicon.ico",
        "update_script": "lib/t_en/update",
        "type": "system"
    },
]


def _import_():
    """ Seed the database with repositories. """
    for repository in repositories:
        repo = Seed(**repository)
        repo.save()



"""
Process actions.
"""
# pylint: disable=too-few-public-methods
from contextlib import suppress
import os
import shutil
from datoso.configuration import config
from datoso.helpers import FileUtils
from datoso.repositories.dedupe import Dedupe
from datoso.database.models.datfile import Dat

class Processor:
    """ Process actions. """
    _previous = None
    actions = []
    seed = None
    file = None

    def __init__(self, **kwargs):
        self._previous = None
        self.__dict__.update(kwargs)

    def process(self):
        """ Process actions. """
        for action in self.actions:
            action_class = globals()[action['action']](file=self.file, seed=self.seed, previous=self._previous, **action)
            yield action_class.process()
            self._previous = action_class.output


class Process:
    """ Process Base class. """
    output = None
    status = None
    previous = {}
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

class LoadDatFile(Process):
    """ Load a dat file. """
    # pylint: disable=no-member
    class_name = None
    file = None
    seed = None
    database = None
    _class = None
    _factory = None
    _dat = None

    def process(self):
        """ Load a dat file. """
        from datoso.database.models.datfile import Dat

        # If there is a factory method, use it to create the class
        if getattr(self, '_factory', None) and self._factory:
            self._class = self._factory(self.file)

        self._dat = self._class(file=self.file)
        self._dat.load()
        self.database = Dat(seed=self.seed, **self._dat.dict())
        self.output = self.database.dict()
        return "Loaded"


class DeleteOld(Process):
    """ Delete old dat file. """
    database = None
    def process(self):
        """ Delete old dat file. """
        from datoso.database.models.datfile import Dat
        self.database = Dat(seed=self.previous['seed'], name=self.previous['name'])
        self.database.load()
        olddat = self.database.dict()
        result = None
        if 'new_file' in olddat and olddat['new_file']:
            try:
                shutil.rmtree(olddat['new_file'])
            except NotADirectoryError:
                with suppress(FileNotFoundError):
                    os.unlink(olddat['new_file'])
            except FileNotFoundError:
                pass
            result = "Deleted"

        self.output = self.previous
        return result


class Copy(Process):
    """ Copy files. """
    destination = None
    file = None
    folder = None
    database = None

    def process(self):
        """ Copy files. """
        from datoso.database.models.datfile import Dat
        origin = self.file if self.file else None
        filename = os.path.basename(origin)
        self.destination = self.destination if self.destination else self.previous['path']

        destination = os.path.join(self.folder, self.destination, filename)
        result = None
        if self.previous:
            self.database = Dat(seed=self.previous['seed'], name=self.previous['name'])
            self.database.load()
            if self.database.is_enabled():
                old_file = self.database.dict().get('new_file', '')
                new_file = destination
                if old_file != new_file or config.getboolean('GENERAL', 'Overwrite', fallback=False) or not os.path.exists(destination):
                    if not old_file:
                        result = "Created"
                    elif old_file != new_file:
                        result = "Updated"
                    elif config.getboolean('GENERAL', 'Overwrite', fallback=False):
                        result = "Overwritten"
                    self.previous['new_file'] = destination
                    FileUtils.copy(origin, destination)
                else:
                    result = "Exists"
            else:
                self.previous['new_file'] = None
                result = "Ignored"
        else:
            FileUtils.copy(origin, destination)
            result = "Copied"

        self.output = self.previous
        return result


class SaveToDatabase(Process):
    """ Save process to database. """
    database = None
    def process(self):
        """ Save process to database. """
        from datoso.database.models.datfile import Dat
        self.database = Dat(**self.previous)
        self.database.save()
        self.database.close()
        self.output = self.previous
        return "Saved"


class Deduplicate(Process):
    """ Save process to database. """
    parent_db = None
    child_db = None

    def get_dat(self, name, seed):
        """ Get dat file. """
        from datoso.database.models.datfile import Dat
        dat = Dat(name=name, seed=seed)
        dat.load()
        return dat

    def process(self):
        """ Save process to database. """
        child_db = Dat(**self.previous)
        child_db.load()
        if not getattr(child_db, 'merge', None) or not getattr(child_db, 'parent', None):
            return "Skipped"
        merged = Dedupe(child_db)
        merged.dedupe()
        merged.save()
        self.output = self.previous
        return "Deduped"


if __name__ == '__main__':
    procesor = Processor()
    procesor.process()



""" Unknown seed, detects version and type of dat already in DatRoot. """
import re
from datoso.repositories.dat import ClrMameProDatFile, XMLDatFile

def detect_xml(dat_file: str, rules):
    """ Detect the seed for a XML dat file. """
    dat = XMLDatFile(file=dat_file)
    for rule_details in rules:
        found = True
        # print(dat.header)
        for rule in rule_details['rules']:
            # print(rule['key'], dat.header.get(rule['key']), rule['value'], rule['operator'], comparator(dat.header.get(rule['key']), rule['value'], rule.get('operator', 'eq')))
            if not comparator(dat.header.get(rule['key']), rule['value'], rule.get('operator', 'eq')):
                found = False
                break
        # print(f'found: {found}')
        if found:
            return rule_details['seed'], rule_details['_class']
    return None, None

def detect_clrmame(dat_file: str, rules):
    """ Detect the seed for a ClrMamePro dat file. """
    dat = ClrMameProDatFile(file=dat_file)
    for rule_details in rules:
        found = True
        # print(dat.header)
        for rule in rule_details['rules']:
            # print(rule['key'], rule['value'], rule['operator'])
            if not comparator(dat.header.get(rule['key']), rule['value'], rule.get('operator', 'eq')):
                found = False
                break
        if found:
            return rule_details['seed'], rule_details['_class']
    return None, None

def detect_seed(dat_file: str, rules):
    """ Detect the seed for a dat file. """
    try:
        return detect_xml(dat_file, rules)
    except Exception:
        try:
            return detect_clrmame(dat_file, rules)
        except Exception:
            return None, None

def comparator(key, value, operator='eq'): # pylint: disable=too-many-return-statements
    """ Returns a boolean based on the comparison of the key and value. """
    match operator:
        case 'eq' | 'equals' | '==':
            return key == value
        case 'ne' | 'not_equals' | '!=':
            return key != value
        case 'gt' | 'greater_than' | '>':
            return key > value
        case 'lt' | 'less_than' | '<':
            return key < value
        case 'ge' | 'greater_than_or_equals' | '>=':
            return key >= value
        case 'le' | 'less_than_or_equals' | '<=':
            return key <= value
        case 'in' | 'is_contained_in' | 'has':
            return key in value if value else False
        case 'ni' | 'not_in' | 'is_not_contained_in' | 'hasnt' | 'has_not':
            return key not in value if value else True
        case 're' | 'regex' | 'matches' | 'match' | 'matches_regex' | 'match_regex':
            return re.search(value, key)
        case 'nr' | 'not_regex' | 'not_matches' | 'not_match' | 'not_matches_regex' | 'not_match_regex':
            return not re.search(value, key)
        case 'sw' | 'starts_with':
            return key.startswith(value) if key else False
        case 'ew' | 'ends_with':
            return key.endswith(value) if key else False
        case 'ns' | 'not_starts_with':
            return not key.startswith(value) if key else True
        case 'ne' | 'not_ends_with':
            return not key.endswith(value) if key else True
        case 'co' | 'contains':
            return value in key if key else False
        case 'nc' | 'not_contains':
            return value not in key if key else True
        case 'ex' | 'exists':
            return bool(key)
        case 'nx' | 'not_exists' | 'not_exist' | 'not_exits' | 'not_exits':
            return not bool(key)
        case 'bt' | 'between' | 'in_range' | 'in_between':
            return value[0] <= key <= value[1] if value else False
        case 'nb' | 'not_between' | 'not_in_range' | 'not_in_between':
            return not (value[0] <= key <= value[1]) if value else True
        case 'is':
            return key is value
        case 'isnt' | 'is_not':
            return key is not value
    return False



"""
Helpers
"""
from contextlib import suppress
import re
import os
from pathlib import Path
from dateutil import parser
import shutil

class Bcolors:
    # pylint: disable=anomalous-backslash-in-string
    """ Color class. """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    BOLDBLUE = '\e[1;34m'
    BOLDCYAN = '\e[1;36m'
    BOLDGREEN = '\e[1;32m'
    BOLDRED = '\e[1;31m'
    BOLDYELLOW = '\e[1;33m'
    BOLDWHITE = '\e[1;37m'
    BOLDMAGENTA = '\e[1;35m'

    @staticmethod
    def color_list() -> list:
        """ Return list of colors. """
        return [attr for attr in Bcolors.__dict__ if not callable(getattr(Bcolors, attr)) and not attr.startswith("__")]

    @staticmethod
    def no_color() -> None:
        """ Disables color output. """
        for color in Bcolors.color_list():
            setattr(Bcolors, color, '')

    @staticmethod
    def remove_color(string) -> str:
        """ Remove color from string. """
        if not string:
            return ''
        for color in Bcolors.color_list():
            string = string.replace(getattr(Bcolors, color), '')
        return string

def is_date(string, fuzzy=False) -> bool:
    """
    Return whether the string can be interpreted as a date.

    :param string: str, string to check for date
    :param fuzzy: bool, ignore unknown tokens in string if True
    """
    try:
        parser.parse(string, fuzzy=fuzzy)
        return True

    except ValueError:
        return False


def sizeof_fmt(num, suffix="B") -> str:
    """ Convert bytes to human readable format. """
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num: .1f}Yi{suffix}"

def is_git_path(path) -> bool:
    """ Check if a path is a git repository. """
    pattern = re.compile((r"((git|ssh|http(s)?)|(git@[\w\.]+))(:(//)?)([\w\.@\:/\-~]+)(\.git)(/)?"))
    return pattern.match(path)

def is_git_repo(path) -> bool:
    """ Check if a path is a git repository. """
    if Path.is_dir(Path(path, ".git")):
        return True
    return False

class FileUtils:

    @staticmethod
    def copy(origin, destination):
        """ Copy file to destination. """
        os.makedirs(os.path.dirname(destination), exist_ok=True)
        try:
            shutil.copy(origin, destination)
        except shutil.SameFileError:
            pass
        except IsADirectoryError:
            with suppress(FileNotFoundError):
                shutil.rmtree(destination)
            shutil.copytree(origin, destination)
        except FileNotFoundError:
            raise FileNotFoundError(f"File {origin} not found.")

    @staticmethod
    def remove(path):
        """ Remove file or folder. """
        try:
            os.unlink(path)
        except IsADirectoryError:
            try:
                shutil.rmtree(path)
            except PermissionError:
                pass
        except FileNotFoundError:
            pass

    @staticmethod
    def parse_folder(path):
        """ Get folder from config. """
        if path is not None and path.startswith(('/', '~')):
            return os.path.expanduser(path)
        else:
            return os.path.join(os.getcwd(), path)

    @staticmethod
    def move(origin, destination):
        """ Move file to destination. """
        os.makedirs(os.path.dirname(destination), exist_ok=True)
        try:
            shutil.move(origin, destination)
        except shutil.Error as error:
            FileUtils.remove(origin)



"""Check if all dependencies are installed"""
from pydoc import locate
from shutil import which
import pkg_resources
from datoso import __app_name__
from datoso.helpers.plugins import installed_seeds
from datoso.helpers import Bcolors


def check_seed(seed):
    """ Check if seed is installed """
    return f'{__app_name__}_seed_{seed}' in installed_seeds()


def check_version(detected, required, expression):
    """ Check if version of required package is correct """
    detected = pkg_resources.parse_version(detected)
    required = pkg_resources.parse_version(required)
    match expression:
        case '>':
            return detected > required
        case '<':
            return detected < required
        case '>=':
            return detected >= required
        case '<=':
            return detected <= required
        case '==':
            return detected == required
        case _:
            return detected == required


def check_module_attributes(seed, module):
    """ Check if all needed files are present """
    reqs = {
        '__preffix__': 'Preffix for identification of dats',
        '__description__': 'Description of module',
        'fetch.fetch': 'Function to fetch data',
        'rules.get_rules': 'Rules for the rules engine',
        'actions.get_actions': 'Actions to take for dats',
        }
    for attribute, desc in reqs.items():
        if not hasattr(module, attribute):
            located = locate(seed+'.'+attribute)
            if located is None:
                print(f'{Bcolors.FAIL} - {Bcolors.BOLD}{seed}.{attribute}{Bcolors.ENDC} not found ({desc})')


def check_executable(executable):
    """ Check if executable is installed """
    return which(executable) is not None or which(executable + '.exe') is not None


def check_module(seed, module, repair=False):
    """ Check if all dependencies are installed """
    if not module:
        module = locate(seed)
    check_module_attributes(seed, module)
    requirements = getattr(module, '__requirements__', None)
    if requirements and 'executables' in requirements:
        for executable in requirements['executables']:
            if not check_executable(executable):
                seed_name = seed.replace(f'{__app_name__}_seed_', '')
                print(f'{seed_name} - {Bcolors.FAIL}{executable}{Bcolors.ENDC} not found.')



"""
Dat classes to parse different types of dat files.
"""
# from abc import abstractmethod
import os
import shlex
import xmltodict

from datoso.database.models.datfile import System


class DatFile:
    """ Base class for dat files. """
    name: str = None
    file: str = None
    full_name: str = None
    seed: str = None

    # calculated values
    modifier: str = None
    system_type: str = None
    company: str = None
    system: str = None
    preffix: str = None
    suffix: str = None
    suffixes = []
    date = None
    path: str = None
    version: str = None

    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)
        if not self.name and not self.file:
            raise ValueError("No file specified")
        if not self.name:
            self.load()

    # @abstractmethod
    def load(self) -> None:
        """ Abstract ** Load the dat file. """

    # @abstractmethod
    def initial_parse(self) -> None:
        """ Parse the dat file. """

    # @abstractmethod
    def get_date(self) -> str:
        """ Get the date from the dat file. """

    def close(self) -> None:
        """ Close the dat file if needed. """

    def get_modifier(self) -> str:
        """ Get the modifier ej. 'Source Code', 'etc' """
        return getattr(self, 'modifier', None)

    def get_company(self) -> str:
        """ Get the company name. """
        return getattr(self, 'company', None)

    def get_system(self) -> str:
        """ Get the system name. """
        return getattr(self, 'system', None)

    def get_system_type(self) -> str:
        """ Get the system type. """
        return getattr(self, 'system_type', None)

    def get_preffix(self) -> str:
        """ Get the preffix for the path. """
        return getattr(self, 'preffix', None)

    def get_suffix(self) -> str:
        """ Get the suffix for the path. """
        return getattr(self, 'suffix', None)

    def get_path(self) -> str:
        """ Get the path for the dat file. """
        self.path = os.path.join(*[x for x in [self.get_preffix(), self.get_company(), self.get_system(), self.get_suffix()] if x])
        return self.path

    def dict(self) -> dict:
        """ Return a dictionary with the dat file information. """
        self.initial_parse()
        return {
            "name": self.name,
            "file": self.file,
            "full_name": self.full_name,
            "date": self.get_date(),
            "modifier": self.get_modifier(),
            "company": self.get_company(),
            "system": self.get_system(),
            "system_type": self.get_system_type(),
            "path": self.get_path(),
        }


class XMLDatFile(DatFile):
    """ XML dat file. """
    shas = None
    game_key = 'game'
    header = None
    merged_roms = []

    def load(self) -> None:
        """ Load the data from a XML file. """
        with open(self.file, encoding='utf-8') as fild:
            self.data = xmltodict.parse(fild.read())
            self.header = self.data['datafile']['header']
            self.name = self.header['name'] if 'name' in self.header else None
            self.full_name = self.header['description'] if 'description' in self.header else None
            self.date = self.header['date'] if 'date' in self.header else None
            self.homepage = self.header['homepage'] if 'homepage' in self.header and self.header['homepage'] and 'insert' not in self.header['homepage'] else None
            self.url = self.header['url'] if 'url' in self.header and self.header['url'] and 'insert' not in self.header['url'] else None
            self.author = self.header['author'] if 'author' in self.header and self.header['author'] and 'insert' not in self.header['author'] else None
            self.email = self.header['email'] if 'email' in self.header and self.header['email'] and 'insert' not in self.header['email'] else None
            self.detect_game_key()

    def save(self) -> None:
        """ Save the data to a XML file. """
        with open(self.file, 'w', encoding='utf-8') as fild:
            fild.write(xmltodict.unparse(self.data, pretty=True))

    def detect_game_key(self) -> str:
        """ Get the game key. """
        for key in self.data['datafile']:
            if key != 'header' and not key.startswith('@'):
                self.game_key = key
                break

    def parse_rom(self, rom: dict) -> dict:
        """ Standarize the rom. """
        parsed_rom = {}
        for key in rom:
            if key.startswith('@'):
                parsed_rom[key[1:]] = rom[key]
            else:
                parsed_rom[key] = rom[key]
        return parsed_rom

    def add_rom(self, rom: dict) -> None:
        """ Add a rom to the dat file. """
        self.shas.add_rom(self.parse_rom(rom))

    def get_rom_shas(self) -> None:
        """ Get the shas for the roms and creates an index. """
        self.shas = HashesIndex()

        for game in self.data['datafile'][self.game_key]:
            if 'rom' not in game:
                continue
            if not isinstance(game['rom'], list):
                self.add_rom(game['rom'])
            else:
                for rom in game['rom']:
                    self.add_rom(rom)

    def merge_with(self, parent: DatFile) -> None:
        """ Merge the dat file with the parent. """
        parent.get_rom_shas()
        for game in self.data['datafile'][self.game_key]:
            if 'rom' not in game:
                continue
            if not isinstance(game['rom'], list):
                if parent.shas.has_rom(self.parse_rom(game['rom'])):
                    self.merged_roms.append(self.parse_rom(game['rom']))
                    del game['rom']
            else:
                new_roms = []
                for rom in game['rom']:
                    if not parent.shas.has_rom(self.parse_rom(rom)):
                        new_roms.append(rom)
                    else:
                        self.merged_roms.append(self.parse_rom(rom))
                game['rom'] = new_roms
        new_games = []
        for game in self.data['datafile'][self.game_key]:
            if 'rom' in game:
                new_games.append(game)
        self.data['datafile'][self.game_key] = new_games

    def get_name(self) -> str:
        """ Get the name of the dat file. """
        if not self.name:
            self.load()
        return self.name

    def overrides(self) -> System:
        """ Overrides data for some systems. """
        find_system = System(company=self.get_company(), system=self.get_system())
        find_system.load()
        if getattr(find_system, 'system_type', None):
            self.system_type = find_system.system_type
            if getattr(find_system, 'override', None):
                self.__dict__.update(find_system.override)
        return find_system

    def extra_configs(self, find_system):
        """ Extra configs for some systems. """
        extra_configs = getattr(find_system, 'extra_configs', None)
        if extra_configs:
            if 'empty_suffix' in extra_configs and not self.suffix:
                self.suffix = extra_configs['empty_suffix'].get(self.seed, None)
            if 'additional_suffix' in extra_configs:
                self.suffix = os.path.join(self.suffix, extra_configs['additional_suffix'].get(self.seed, None))
            if 'if_suffix' in extra_configs:
                for key, value in extra_configs['if_suffix'].items():
                    if key in self.suffixes:
                        self.__dict__.update(value)

    def get_date(self):
        """ Get the date from the dat file. """
        return self.date

class ClrMameProDatFile(DatFile):
    """ ClrMamePro dat file. """
    header = {}
    games = []

    def get_next_block(self, data):
        """ Get the next block of data. """
        parenthesis = 0
        start = 0
        end = 0
        for i, char in enumerate(data):
            if char == '(':
                if parenthesis == 0:
                    start = i + 1
                parenthesis += 1
            if char == ')':
                parenthesis -= 1
            if parenthesis == 0 and start >= 1:
                end = i
                break
        return data[start:end], data[end + 1:] if end < len(data) else None

    def read_block(self, data) -> dict:
        """ Read a block of data from a ClrMame dat and parses it. """
        dictionary = {}
        for line in iter(data.splitlines()):
            line = line.strip()
            if line:
                if line.startswith('rom'):
                    line = line[6:-2]
                    rom = {'@name': None, '@crc': None, '@md5': None, '@sha1': None}
                    try:
                        data = shlex.split(line)
                    except ValueError:
                        data = line.split(' ')
                    for i in range(0, len(data), 2):
                        rom[f'@{data[i]}'] = data[i+1]
                    dictionary['rom'] = [] if 'rom' not in dictionary else dictionary['rom']
                    dictionary['rom'].append(rom)
                else:
                    try:
                        key, value = shlex.split(line)
                    except ValueError as exc:
                        raise ValueError(f'Error parsing line: {line} from: {self.file}') from exc
                    dictionary[key] = value

        return dictionary

    def load(self) -> None:
        """ Load the data from a ClrMamePro file. """
        self.games = []
        with open(self.file, encoding='utf-8', errors='ignore') as file:
            data = file.read()

            block, next_block = self.get_next_block(data)
            self.header = self.read_block(block)

            while next_block:
                block, next_block = self.get_next_block(next_block)
                self.games.append(self.read_block(block))

        self.data = {
            'datafile': {
                'header':  self.header,
                'game': self.games
            }
        }
        self.name = self.header['name']
        self.full_name = self.header['description']

    def get_rom_shas(self) -> None:
        """ TODO Method """


class ZipMultiDatFile(DatFile):
    """ Base class for dat files. """
    def get_header(self) -> dict:
        """ Get the header from the dat file. """
        return { 'name': os.path.basename(self.file), 'description': self.file }

    def load(self) -> None:
        """ Load the data from a ClrMamePro file. """
        self.games = []
        self.header = self.get_header()

        self.data = {
            'datafile': {
                'header':  self.header,
                'game': self.games
            }
        }
        self.name = self.header['name']
        self.full_name = self.header['description']

class DirMultiDatFile(DatFile):
    """ Base class for dat files. """
    def get_header(self) -> dict:
        """ Get the header from the dat file. """
        return { 'name': self.name, 'description': self.file }

    def load(self) -> None:
        """ Load the data from a ClrMamePro file. """
        self.games = []
        for file in os.listdir(self.file):
            if file.endswith(('.xml', '.dat')):
                self.games.append({ 'rom': { '@name': file } })

        self.name = os.path.basename(self.file)

        self.header = self.get_header()

        self.data = {
            'datafile': {
                'header':  self.header,
                'game': self.games
            }
        }
        self.name = self.header['name']
        self.full_name = self.header['description']


class HashesIndex:
    """ Index of hashes. """
    hashes = ['sha256', 'sha1', 'md5', 'crc']
    sha256 = {}
    sha1 = {}
    md5 = {}
    crc = {}
    sizes = {}

    def add_rom(self, rom):
        """ Add a rom to the index. """
        for hash in self.hashes:
            if hash in rom:
                if hash not in self.__dict__:
                    self.__dict__[hash] = {}
                self.__dict__[hash][rom[hash]] = rom

    def has_rom(self, rom, hash=None):
        """ Check if a rom exists in the index. """
        if hash:
            return hash in rom and hash in self.__dict__ \
                and rom[hash] in self.__dict__[hash] \
                and rom['size'] == self.__dict__[hash][rom[hash]]['size']
        return any((hash in rom and hash in self.__dict__ \
                    and rom[hash] in self.__dict__[hash] \
                    and rom['size'] == self.__dict__[hash][rom[hash]]['size']) \
                    for hash in self.hashes)

    def get_sha256s(self):
        """ Get the sha256s. """
        return self.sha256.keys()

    def get_sha1s(self):
        """ Get the sha1s. """
        return self.sha1.keys()

    def get_md5s(self):
        """ Get the md5s. """
        return self.md5.keys()

    def get_crcs(self):
        """ Get the crcs. """
        return self.crc.keys()



"""Main entry point for datoso"""
import configparser
import json
import logging
import os
from pathlib import Path
import re
import sys
import argparse
from venv import logger
from tabulate import tabulate
from datoso.configuration.configuration import get_seed_name
from datoso.configuration.logger import enable_logging, set_verbosity
from datoso.database.models.datfile import Dat

from datoso import __version__, __app_name__, ROOT_FOLDER
from datoso.helpers import Bcolors
from datoso.configuration import config

from datoso.helpers.plugins import installed_seeds, seed_description
from datoso.commands.doctor import check_module, check_seed
from datoso.commands.seed import Seed
from datoso.repositories.dedupe import Dedupe
from datoso.seeds.rules import Rules
from datoso.seeds.unknown_seed import detect_seed

#---------Boilerplate to check python version ----------
if sys.version_info[0] < 3 or sys.version_info.minor < 9:
    print("This is a Python 3 script. Please run it with Python 3.9 or above")
    sys.exit(1)


def parse_args() -> argparse.Namespace:
    """Parse command line arguments

    Returns:
        argparse.Namespace: An object to take the attributes.
    """
    #pylint: disable=too-many-locals,too-many-statements
    parser = argparse.ArgumentParser(description='Update dats from different sources.')
    subparser = parser.add_subparsers(help='sub-command help')

    parser.add_argument('-v', '--version', action='store_true', help='show version')

    parser_save = subparser.add_parser('config', help='Show configuration')
    parser_save.add_argument('-s', '--save', action='store_true', help='Save configuration to .datosorc')
    parser_save.add_argument('-d', '--directory', default='~', choices=['~', '.'], help='Directory to save .datosorc')
    parser_save.set_defaults(func=command_config)
    parser_save.add_argument('-ru', '--rules-update', action='store_true', help='Update system rules from GoogleSheets Url')

    group_save = parser_save.add_mutually_exclusive_group()
    group_save.add_argument('--set', nargs=2, metavar=('configuration', 'value'), help='Set Configuration Option separated by point with new value e.g. <GENERAL.Overwrite> <false>')
    group_save.add_argument('--get', metavar=('configuration'), help='Get value of Configuration Option.')
    parser_save.add_argument('-g','--global', action='store_true', help='When set, saves to global config, else to `.datosorc`')

    parser_doctor = subparser.add_parser('doctor', help='Doctor installed seeds')
    parser_doctor.add_argument('seed', nargs='?', help='Seed to doctor')
    parser_doctor.set_defaults(func=command_doctor)
    parser_doctor.add_argument('-r', '--repair', action='store_true', help='Try to repair seed(s)')

    parser_dat = subparser.add_parser('dat', help='Changes configuration in current dats')
    parser_dat.add_argument('command', nargs='?', help='Command to execute')

    group_dat= parser_dat.add_mutually_exclusive_group(required=True)
    group_dat.add_argument('-d', '--dat-name', help='Select dat to update/check, must be in format "seed:name"')
    group_dat.add_argument('-f', '--find', help='Select dats based on filter, they are "<field><operator><value>;...", valid operators are: =, !=, and ~=')
    group_dat.add_argument('-a', '--all', help='Show all dats', action='store_true')

    parser_dat.add_argument('-s', '--set', help='Manually set variable, must be in format "variable=value"')
    parser_dat.add_argument('--delete', action='store_true', default=False, help='Delete Dat')

    parser_dat.add_argument('-on', '--only-names', action='store_true', help='Only show names')

    parser_dat.set_defaults(func=command_dat)

    # Seed admin commands
    parser_seed = subparser.add_parser('seed', help='Seed admin commands')
    subparser_seed = parser_seed.add_subparsers(help='sub-command help')

    parser_available = subparser_seed.add_parser('installed', help='List installed seeds')
    parser_available.set_defaults(func=command_seed_installed)

    parser_details = subparser_seed.add_parser('details', help='Show details of seed')
    parser_details.add_argument('seed', help='Seed to show details of')
    parser_details.set_defaults(func=command_seed_details)

    parser_import = subparser.add_parser('import', help='Import dats from existing romvault')
    parser_import.set_defaults(func=command_import)

    parser_deduper = subparser.add_parser('deduper', help='Deduplicate dats, removes duplicates from input dat existing in parent dat')
    parser_deduper.add_argument('-i', '--input', required=True, help='Input dat file e.g. "redump:psx_child" or "/mnt/roms/redump_psx_child.dat"')
    parser_deduper.add_argument('-p', '--parent', default=None, help='Parent dat file e.g. "redump:psx" or "/mnt/roms/redump_psx.dat" if not set, parent is taken from input dat with prescanned dats')
    parser_deduper.add_argument('-o', '--output', default=None, help='If different from input.dat, must be a file path e.g. "/mnt/roms/redump_psx_child_deduped.dat"')
    parser_deduper.add_argument('-dr', '--dry-run', action='store_true', help='Do not write output file, just show actions')

    parser_deduper.set_defaults(func=command_deduper)


    # Seed commands
    for seed, _ in list(installed_seeds().items()) + [('all', 'All seeds')]:
        seed = get_seed_name(seed)
        parser_command = subparser.add_parser(seed, help=f'Update seed {seed}')
        parser_command.set_defaults(func=command_seed, seed=seed)
        parser_command.add_argument('-f', '--fetch', action='store_true', help='Fetch seed')
        parser_command.add_argument('-p', '--process', action='store_true', help='Process dats from seed')
        parser_command.add_argument('-d', '--details', action='store_true', help='Show details of seed')
        parser_command.add_argument('-fd', '--filter', help='Filter dats to process')
        if seed == 'all':
            parser_command.add_argument('-e', '--exclude', action='append', help='Exclude seed or seeds (only work with all)')
            parser_command.add_argument('-o', '--only', action='append', help='Only seed or seeds (only work with all)')

    # Common arguments
    subparsers = [subparser, subparser_seed]
    for subpars in subparsers:
        for subpar in subpars.choices.values():
            subpar.add_argument('-v', '--verbose', action='store_true', help='verbose output')
            subpar.add_argument('-q', '--quiet', action='store_true', help='quiet output')
            subpar.add_argument('-nc', '--no-color', action='store_true', help='disable color output')
            subpar.add_argument('-l', '--logging', action='store_true', help='enable logging')

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args()
    initial_setup(args)
    return args


def initial_setup(args) -> None:
    """ Initial setup of datoso from command line arguments """
    if getattr(args, 'version', False):
        print(__version__)
        sys.exit()
    if getattr(args, 'no_color', False) or os.name == 'nt':
        Bcolors.no_color()
    if getattr(args, 'quiet', False):
        set_verbosity(logging.WARNING)
        config['COMMAND']['Quiet'] = 'true'
    if getattr(args, 'verbose', False):
        set_verbosity(logging.DEBUG)
        config['COMMAND']['Verbose'] = 'true'
    if getattr(args, 'logging', False):
        enable_logging()


def command_deduper(args) -> None:
    """ Deduplicate dats, removes duplicates from input dat existing in parent dat """
    if not args.parent and args.input.endswith(('.dat', '.xml')):
        print('Parent dat is required when input is a dat file')
        sys.exit(1)
    if args.dry_run:
        logger.setLevel(logging.DEBUG)
    merged = Dedupe(args.input, args.parent)
    merged.dedupe()
    if args.output and not args.dry_run:
        merged.save(args.output)
    elif not args.dry_run:
        merged.save(args.input)
    print(f'{Bcolors.OKBLUE}File saved to {args.output if args.output else args.input}{Bcolors.ENDC}')


def command_import(_) -> None:
    """ Make changes in dat config """
    dat_root_path = config['PATHS']['DatPath']

    if not dat_root_path or not Path(dat_root_path).exists():
        print(f'{Bcolors.FAIL}Dat root path not set or does not exist{Bcolors.ENDC}')
        sys.exit(1)

    rules = Rules().rules

    dats = { str(x):None for x in Path(dat_root_path).rglob("*.[dD][aA][tT]") }

    if config['IMPORT'].get('IgnoreRegEx'):
        ignore_regex = re.compile(config['IMPORT']['IgnoreRegEx'])
        dats = [ dat for dat in dats if not ignore_regex.match(dat) ]

    fromhere = ''
    found = False
    for dat_name in dats:
        if dat_name == fromhere or fromhere == '':
            found = True
        if not found:
            continue
        print(f'{dat_name} - ', end='')
        seed, _class = detect_seed(dat_name, rules)
        print(f'{seed} - {_class.__name__ if _class else None}')
        if _class:
            dat = _class(file=dat_name)
            dat.load()
            database = Dat(seed=seed, new_file=dat_name, **dat.dict())
            database.save()
            database.close()


def command_dat(args):
    """ Make changes in dat config """

    def print_dats(dats):
        """ Print dats """
        output = []
        for dat in dats:
            output.append({
                'seed': dat['seed'],
                'name': dat['name'],
                'status': dat['status'] if 'status' in dat else 'enabled',
            })
        if getattr(args, 'only_names', False):
            for dat in output:
                print(f"{dat['seed']}:{dat['name']}")
            return
        print(tabulate(output, headers='keys', tablefmt='psql'))

    from datoso.database import DB
    from tinydb import Query
    query = Query()
    table = DB.table('dats')
    if args.dat_name:
        splitted = args.dat_name.split(':')
        if len(splitted) != 2:
            print(f'{Bcolors.WARNING}Invalid dat name, must be in format "seed:name"{Bcolors.ENDC}')
            print(f'Showing results for filter: {Bcolors.OKCYAN}name~={args.dat_name}{Bcolors.ENDC}')
            print('--------------------------------------------------------------')
            name = args.dat_name
            result = table.search(query.name.matches(r'^.*' + name + r'.*', flags=re.IGNORECASE))

            print_dats(result)

        else:
            seed, name = splitted
            result = table.get((query.name == name) & (query.seed == seed))
            if not result:
                print(f'{Bcolors.FAIL}Dat not found{Bcolors.ENDC}')
                sys.exit(1)
            if args.set:
                key, value = args.set.split('=') if '=' in args.set else (args.set, True)
                if value.isdigit():
                    value = int(value)
                if value.lower() == 'true':
                    value = True
                table.update({key: value}, doc_ids=[result.doc_id])
                table.storage.flush()
                print(f'{Bcolors.OKGREEN}Dat {Bcolors.OKCYAN}{seed}:{name}{Bcolors.OKGREEN} {key} set to {Bcolors.OKBLUE}{value}{Bcolors.ENDC}')
                sys.exit(0)
            if args.delete:
                table.remove(doc_ids=[result.doc_id])
                table.storage.flush()
                print(f'{Bcolors.OKGREEN}Dat {Bcolors.OKCYAN}{seed}:{name}{Bcolors.OKGREEN} removed{Bcolors.ENDC}')
                sys.exit(0)
            print_dats([result])
    elif args.all:
        # Show all dats
        print_dats(table.all())

    elif args.find:
        # Find dats TODO: finish
        print(f'Showing results for filter: {Bcolors.OKCYAN}{args.find}{Bcolors.ENDC}')
        print('--------------------------------------------------------------')
        name, value = args.find.split('=')
        from tinydb import where
        result = table.search(where(name) == value)
        print_dats(result)


def command_seed_installed(_) -> None:
    """ List available seeds """
    print('Installed seeds:')
    for seed, seed_module in installed_seeds().items():
        description = seed_description(seed_module)
        description = {description[0:60]+'...' if len(description) > 60 else description}
        seed_name = seed[12:]
        print(f'* {Bcolors.OKGREEN}{seed_name}{Bcolors.ENDC} - {description}')


def command_seed_details(args) -> None:
    module = None
    for seed, seed_module in installed_seeds().items():
        if seed == f'{__app_name__}_seed_{args.seed}':
            module = seed_module
            break
    if not module:
        print(f'Seed {Bcolors.FAIL}{args.seed}{Bcolors.ENDC} not installed')
        sys.exit(1)
    print(f'Seed {Bcolors.OKGREEN}{args.seed}{Bcolors.ENDC} details:')
    print(f'  * Name: {module.__name__}')
    print(f'  * Version: {module.__version__}')
    print(f'  * Author: {module.__author__}')
    print(f'  * Description: {module.__description__}')


def command_seed(args) -> None:
    """ Commands with the seed (must be installed) """
    if args.seed == 'all':
        for seed, _ in installed_seeds().items():
            seed = get_seed_name(seed)
            if seed in args.exclude:
                continue
            if args.only and seed not in args.only:
                continue
            if config['PROCESS'].get('SeedIgnoreRegEx'):
                ignore_regex = re.compile(config['PROCESS']['SeedIgnoreRegEx'])
                if ignore_regex.match(seed):
                    continue
            args.seed = seed
            command_seed(args)
        sys.exit(0)
    seed = Seed(name=args.seed)
    if getattr(args, 'details', False):
        command_seed_details(args)
    if getattr(args, 'fetch', False):
        print(f'Fetching seed {Bcolors.OKGREEN}{args.seed}{Bcolors.ENDC}')
        if seed.fetch():
            print(f'Errors fetching {Bcolors.FAIL}{args.seed}{Bcolors.ENDC}')
            print('Please enable logs for more information or use -v parameter')
            command_doctor(args)
    if getattr(args, 'process', False):
        print('=======================')
        print(f'{Bcolors.OKCYAN}Processing seed {Bcolors.OKGREEN}{args.seed}{Bcolors.ENDC}')
        print('=======================')
        if seed.process_dats(fltr=getattr(args, 'filter', None)):
            print(f'Errors processing {Bcolors.FAIL}{args.seed}{Bcolors.ENDC}')
            print('Please enable logs for more information or use -v parameter')
            command_doctor(args)


def command_config_save(args) -> None:
    """ Save config to file """
    config_file = os.path.expanduser('~/.datosorc') if args.directory == '~' else os.path.join(os.getcwd(), '.datosorc')
    with open(config_file, 'w', encoding='utf-8') as file:
        config.write(file)
    print(f'Config saved to {Bcolors.OKGREEN}{config_file}{Bcolors.ENDC}')


def command_config_set(args) -> None:
    """ Set config value, if global is set, it will be set in datoso.ini file """
    myconfig = args.set[0].split('.')
    if len(myconfig) != 2:
        print(f'{Bcolors.FAIL}Invalid config key, must be in <SECTION>.<Option> format. {Bcolors.ENDC}')
        sys.exit(1)
    if myconfig[1] not in config[myconfig[0]]:
        print(f'{Bcolors.FAIL}Invalid config option. {Bcolors.ENDC}')
        sys.exit(1)

    newconfig = configparser.ConfigParser(comment_prefixes='/', allow_no_value=True)
    newconfig.optionxform = lambda option: option
    if getattr(args, 'global', False):
        file = os.path.join(ROOT_FOLDER, 'datoso.ini')
    else:
        file = os.path.join(os.getcwd(), '.datosorc')
        if not os.path.isfile(file):
            file = os.path.expanduser('~/.datosorc')
    newconfig.read(file)
    if not newconfig.has_section(myconfig[0]):
        newconfig.add_section(myconfig[0])
    newconfig[myconfig[0]][myconfig[1]] = args.set[1]
    with open(file, 'w', encoding='utf-8') as file:
        newconfig.write(file)
    if getattr(args, 'global', False):
        print(f'{Bcolors.OKGREEN}Global config {Bcolors.OKCYAN}{myconfig[0]}.{myconfig[1]}{Bcolors.OKGREEN} set to {Bcolors.OKBLUE}{args.set[1]}{Bcolors.ENDC}')
    else:
        print(f'{Bcolors.OKGREEN}Local Config {Bcolors.OKCYAN}{myconfig[0]}.{myconfig[1]}{Bcolors.OKGREEN} set to {Bcolors.OKBLUE}{args.set[1]}{Bcolors.ENDC}')


def command_config_get(args) -> None:
    """ Get active config value """
    myconfig = args.get.split('.')
    if len(myconfig) != 2:
        print(myconfig)
        print(f'{Bcolors.FAIL}Invalid config key, must be in <SECTION>.<Option> format. {Bcolors.ENDC}')
        sys.exit(1)
    if myconfig[1] not in config[myconfig[0]]:
        print(f'{Bcolors.FAIL}Invalid config option. {Bcolors.ENDC}')
        sys.exit(1)
    print(config[myconfig[0]][myconfig[1]])


def command_config_rules_update(args) -> None:
    """ Update rules from google sheet """
    from datoso.database.seeds import dat_rules
    print('Updating rules')
    try:
        dat_rules.import_dats()
        print('Rules updated')
    except Exception as exc:
        print(f'{Bcolors.FAIL}Error updating rules{Bcolors.ENDC}')
        print(exc)
        print('Please enable logs for more information or use -v parameter')
        command_doctor(args)


def command_config(args) -> None:
    """ Config commands """
    if args.save:
        command_config_save(args)
    elif args.set:
        command_config_set(args)
    elif args.get:
        command_config_get(args)
    elif args.rules_update:
        command_config_rules_update(args)
    else:
        config_dict = {s:dict(config.items(s)) for s in config.sections()}
        print(json.dumps(config_dict, indent=4))


def command_list(_):
    """ List installed seeds """
    for seed, seed_class in installed_seeds().items():
        description = seed_class.description()
        print(f'* {Bcolors.OKCYAN}{seed}{Bcolors.ENDC} - {description[0:60] if len(description) > 60 else description}...')

def command_doctor(args):
    """ Doctor installed seeds """
    if getattr(args, 'seed', False):
        seed = check_seed(args.seed)
        if not seed:
            print(f'Module Seed {Bcolors.FAIL}  - {Bcolors.BOLD}{args.seed}{Bcolors.ENDC} not found')
            sys.exit(1)
        seed = seed if args.seed.startswith('datoso_seed_') else f'datoso_seed_{args.seed}'
        seeds = {seed: None}
    else:
        seeds = installed_seeds()
    for seed, module in seeds.items():
        print(f'Checking seed {Bcolors.OKCYAN}{seed}{Bcolors.ENDC}')
        check_module(seed, module, args.repair)

def main():
    """ Main function """
    from datoso.database.seeds.dat_rules import detect_first_run
    detect_first_run()
    args = parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



import logging
from datoso.database.models.datfile import Dat
from datoso.repositories.dat import DatFile, XMLDatFile, ClrMameProDatFile

class Dedupe:
    """ Merge two dat files. """

    child = {}
    parent = {}

    def __init__(self, child, parent=None):
        self.child = {}
        self.parent = {}

        def load_metadata(var, obj):
            if isinstance(var, str):
                if ':' in var:
                    splitted = var.split(':')
                    seed = splitted[0]
                    name = splitted[1]
                    obj['db'] = Dat(name=name, seed=seed)
                    obj['db'].load()
                    var = obj['db']
                elif var.endswith(('.dat', '.xml')):
                    obj['file'] = var
                else:
                    raise Exception("Invalid dat file")
            if isinstance(var, Dat):
                obj['db'] = var
                obj['file'] = getattr(var, 'new_file', None) or var.file
            if isinstance(var, DatFile):
                obj['dat'] = var
            else:
                obj['dat'] = self.get_dat_file(obj['file'])

        load_metadata(child, self.child)
        if not parent:
            parent = self.child['db'].parent
        load_metadata(parent, self.parent)

    def get_dat_file(self, file):
        try:
            dat = XMLDatFile(file=file)
            dat.load()
            return dat
        except Exception:
            pass
        try:
            dat = ClrMameProDatFile(file=file)
            dat.load()
            return dat
        except Exception:
            pass
        raise Exception("Invalid dat file")

    def dedupe(self):
        self.child['dat'].merge_with(self.parent['dat'])
        logging.info("Deduped %i roms", len(self.child['dat'].merged_roms))
        return self.child['dat']

    def save(self, file=None):
        if file:
            self.child['dat'].file = file
        self.child['dat'].save()


""" Configuration module """
__all__ = ['config', 'logger', 'ROOT_FOLDER']

from datoso import ROOT_FOLDER

from datoso.configuration.configuration import config
from datoso.configuration.logger import logger



"""
    Seed the database with Systems.
"""
import json
import requests
from pathlib import Path
from datoso.configuration import config
from datoso import ROOT_FOLDER
from datoso.database.models import System


fields = [
    'company',
    'system',
    'override',
    'extra_configs',
    'system_type',
]


def get_systems():
    """ Get systems from the Google Sheets. """
    if not config['UPDATE_URLS']['GoogleSheetUrl']:
        return []
    result = requests.get(config['UPDATE_URLS']['GoogleSheetUrl'], timeout=300)
    systems_result = result.json()['values']
    systems = []
    for i in range(1, len(systems_result)):
        system = systems_result[i]
        row = {}
        for j, field in enumerate(fields):
            if len(system) > j and system[j] != '':
                row[field] = system[j]
            elif field == 'company':
                row[field] = None
            if field in ('override', 'extra_configs') and field in row:
                try:
                    row[field] = json.loads(system[j])
                except Exception:  # pylint: disable=broad-except
                    row[field] = system[j]
        systems.append(row)
    return systems


def import_dats():
    """ Seed the database with Systems. """
    # pylint: disable=protected-access
    systems = get_systems()
    with open(Path(ROOT_FOLDER,'systems.json'), 'w', encoding='utf-8') as file:
        json.dump(systems, file, indent=4)
    for system in systems:
        row = System(**system)
        row.save()
        row._DB.table.storage.flush()

def init():
    """ Seed the database with Systems. """
    # pylint: disable=protected-access
    with open(Path(ROOT_FOLDER,'systems.json'), 'r', encoding='utf-8') as file:
        systems = json.loads(file.read())
    for system in systems:
        row = System(**system)
        row.save()
        row._DB.table.storage.flush()

def detect_first_run():
    """ Detect if this is the first run. """
    # pylint: disable=protected-access
    if not len(System.all()):
        init()